class RestMethod:
    DELETE = 'DELETE'
    GET = 'GET'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    POST = 'POST'
    PUT = 'PUT'
    TRACE = 'TRACE'
    ALL = [DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT, TRACE]  # noqa: RUF012
